import random
import math
import numpy as np
import time


class Utilities:
    """Simple utilities to display time
    """

    @staticmethod
    def milliseconds_to_table(t):
        """Convert time in milliseconds to human-friendly format

        :param t: A nonnegative integer. Time in milliseconds.

        :return: [y, mo, d, h, mi, s, ms]. Such that t milliseconds =
        y years + mo months + d days + h hours + mi minutes + s seconds +
        ms milliseconds (rounded to nearest millisecond). By convention,
        1 year = 365 days and 1 month = 31 days.
        """
        time_left = int(round(t, 0))                # time_left is in ms
        (time_left, ms) = divmod(time_left, 1000)   # time_left is now in s
        (time_left, s) = divmod(time_left, 60)      # time_left is now in mn
        (time_left, mi) = divmod(time_left, 60)     # time_left is now in hours
        (time_left, h) = divmod(time_left, 24)      # time_left is now in days
        (y, days_in_year) = divmod(time_left, 365)
        (mo, d) = divmod(days_in_year, 31)
        return [y, mo, d, h, mi, s, ms]

    @staticmethod
    def int_and_unit_to_string(number, unit):
        """Convert an integer and a unit to a string

        :param number: A nonnegative integer.
        :param unit: A string. Name of the unit (in singular form). The plural
        form is assumed to be unit + "s".

        :return: A string. The number and the unit, with final "s" if plural.
        """
        if number <= 1:
            return str(number) + " " + unit
        else:
            return str(number) + " " + unit + "s"

    @staticmethod
    def milliseconds_to_string(t):
        """Convert an integer and a unit to a string

        :param t: A nonnegative integer. Time in milliseconds.

        :return: A string of the form "y years, mo months, d days, h hours,
        mi minutes, s seconds and ms milliseconds".
        """
        [y, mo, d, h, mi, s, ms] = Utilities.milliseconds_to_table(t)

        # Simple implementation:
        # return (
        #     Utilities.int_and_unit_to_string(y, "year") + ", " +
        #     Utilities.int_and_unit_to_string(mo, "month") + ", " +
        #     Utilities.int_and_unit_to_string(d, "day") + ", " +
        #     Utilities.int_and_unit_to_string(h, "hour") + ", " +
        #     Utilities.int_and_unit_to_string(mi, "minute") + ", " +
        #     Utilities.int_and_unit_to_string(s, "second") + " and " +
        #     Utilities.int_and_unit_to_string(ms, "millisecond")
        # )

        # Nicer version:
        result = []
        if y > 0:
            result.append(Utilities.int_and_unit_to_string(y, "year"))
        if mo > 0:
            result.append(Utilities.int_and_unit_to_string(mo, "month"))
        if d > 0:
            result.append(Utilities.int_and_unit_to_string(d, "day"))
        if h > 0:
            result.append(Utilities.int_and_unit_to_string(h, "hour"))
        if mi > 0:
            result.append(Utilities.int_and_unit_to_string(mi, "minute"))
        if s > 0:
            result.append(Utilities.int_and_unit_to_string(s, "second"))
        if ms > 0:
            result.append(Utilities.int_and_unit_to_string(ms, "millisecond"))
        l = len(result)
        if l == 0:
            return "0 millisecond"
        elif l == 1:
            return result[0]
        else:
            result_string = result[0]
            for word in result[1:len(result)-1]:
                result_string = result_string + ", " + word
            result_string = result_string + " and " + result[len(result)-1]
            return result_string


def random_table(table_length):
    """Table of random integers between 1 and 1000

    :param table_length: A nonnegative integer. Length of the table.

    :return: A table.
    """
    return np.random.randint(1000, size=table_length) + 1


class Tri:
    """Sorting functions
    """

    @staticmethod
    def insertion_sort(tab):
        """Sort a table by insertion.

        :param tab: A table of numbers.

        :return: The table, sorted in place.
        """
        l = len(tab)
        for i in range(l):
            # i : initial position of element to sort
            j = i
            # j : current position of element to sort
            while j > 0 and tab[j-1] > tab[j]:
                (tab[j], tab[j-1]) = (tab[j-1], tab[j])
                j -= 1
        return tab

    @staticmethod
    def bubble_sort(tab):
        """Sort a table, using bubble sort.

        :param tab: A table of numbers.

        :return: The table, sorted in place.
        """
        i = len(tab)  # i : final position of element moving to "surface"
        exchange_done = True
        while i > 0 and exchange_done:
            exchange_done = False
            for j in range(0, i-1):
                # j : position of element that might move
                if tab[j] > tab[j+1]:
                    (tab[j], tab[j+1]) = (tab[j+1], tab[j])
                    exchange_done = True
            i -= 1
        return tab

    @staticmethod
    def merge_tables(tab1, tab2):
        """Merge two sorted tables.

        :param tab1: A table of numbers, sorted in ascending order.
        :param tab2: A table of numbers, sorted in ascending order.

        :return: The merged table, sorted in ascending order.
        """
        result = []
        i = 0
        j = 0
        while i < len(tab1) and j < len(tab1):
            if tab1[i] <= tab2[j]:
                result.append(tab1[i])
                i += 1
            else:
                result.append(tab2[j])
                j += 1

        if tab1:
            result.extend(tab1[i:])
        if tab2:
            result.extend(tab2[j:])
        return result

    @staticmethod
    def merge_sort(tab):
        """Sort a table, using merge sort.

        :param tab: A table of numbers.

        :return: The table, sorted in ascending order.
        """
        if len(tab) <= 1:
            return tab

        middle = len(tab) // 2
        tab1 = tab[:middle]
        tab2 = tab[middle:]

        tab1 = Tri.merge_sort(tab1)
        tab2 = Tri.merge_sort(tab2)
        return list(Tri.merge_tables(tab1, tab2))

    @staticmethod
    def quick_sort(tab):
        """Sort a table, using quick sort.

        :param tab: A table of numbers.

        :return: The table, sorted in ascending order.
        """
        if len(tab) <= 1:
            return tab

        index_pivot = random.randint(0, len(tab)-1)
        pivot = tab[index_pivot]
        less = []
        equal = []
        more = []
        for element in tab:
            if element < pivot:
                less.append(element)
            elif element > pivot:
                more.append(element)
            else:
                equal.append(element)
        less = Tri.quick_sort(less)
        more = Tri.quick_sort(more)
        return less + equal + more

if __name__ == '__main__':

    print("Exercise 1")

    t_milliseconds = random.randint(0, 10**12)
    print("")
    print("t_milliseconds = ")
    print(t_milliseconds)

    t_table = Utilities.milliseconds_to_table(t_milliseconds)
    print("")
    print("t_table = ")
    print(t_table)

    (years, months, days, hours, minutes, seconds, milliseconds) = t_table
    t_check = (
        (
            (
                (years * 365 + months * 31 + days) * 24 + hours
            ) * 60 + minutes
        ) * 60 + seconds
    ) * 1000 + milliseconds
    print("")
    print("Sanity check:")
    print("t_check =")
    print(t_check)

    print("")
    print("t (user friendly) = ")
    print(Utilities.milliseconds_to_string(t_milliseconds))

    n = 100
    print("")
    print("n = ", n)
    print("")
    print("log(n) = ")
    execution_time = int(math.log(n))
    print(Utilities.milliseconds_to_string(execution_time))
    print("")
    print("n = ")
    execution_time = int(n)
    print(Utilities.milliseconds_to_string(execution_time))
    print("")
    print("n ln(n) = ")
    execution_time = int(n * math.log(n))
    print(Utilities.milliseconds_to_string(execution_time))
    print("")
    print("n^2 = ")
    execution_time = int(n ** 2)
    print(Utilities.milliseconds_to_string(execution_time))
    print("")
    print("n^3 = ")
    execution_time = int(n ** 3)
    print(Utilities.milliseconds_to_string(execution_time))
    print("")
    print("2^n = ")
    execution_time = int(2 ** n)
    print(Utilities.milliseconds_to_string(execution_time))

    print("")
    print("Exercise 2")

    print("")
    print("Sanity checks")
    tab_unsorted = random_table(10)
    print(tab_unsorted)
    print(Tri.insertion_sort(np.copy(tab_unsorted)))
    print(Tri.bubble_sort(np.copy(tab_unsorted)))
    print(np.array(Tri.merge_sort(np.copy(tab_unsorted))))
    print(np.array(Tri.quick_sort(np.copy(tab_unsorted))))

    print("")
    print("Duration tests")
    tab_unsorted = random_table(1000)

    print("Insertion sort")
    time1 = time.time()
    tab_insertion = Tri.insertion_sort(np.copy(tab_unsorted))
    time2 = time.time()
    print(Utilities.milliseconds_to_string(1000*(time2 - time1)))

    print("Bubble sort")
    time1 = time.time()
    tab_bubble = Tri.bubble_sort(np.copy(tab_unsorted))
    time2 = time.time()
    print(Utilities.milliseconds_to_string(1000*(time2 - time1)))

    print("Merge sort")
    time1 = time.time()
    tab_merge = Tri.merge_sort(np.copy(tab_unsorted))
    time2 = time.time()
    print(Utilities.milliseconds_to_string(1000*(time2 - time1)))

    print("Quick sort")
    time1 = time.time()
    tab_quick = Tri.quick_sort(np.copy(tab_unsorted))
    time2 = time.time()
    print(Utilities.milliseconds_to_string(1000*(time2 - time1)))