def fibo_rec(n):
    """Fibonacci sequence (recursive computation)

    :param n: A nonnegative integer.

    :return: f_n, the n-th Fibonacci number, with f_0 = 0, f_1 = 1 and
    f_n = f_{n-2} + f_{n-1} for n >= 2.
    """
    if n == 0:
        return 0
    if n == 1:
        return 1
    return fibo_rec(n - 2) + fibo_rec(n - 1)


def fibo_it(n):
    """Fibonacci sequence (iterative computation)

    :param n: A nonnegative integer.

    :return: f_n, the n-th Fibonacci number, with f_0 = 0, f_1 = 1 and
    f_n = f_{n-2} + f_{n-1} for n >= 2.
    """
    if n == 0:
        return 0
    x = 0
    y = 1
    for i in range(2, n+1):
        # Beginning of loop :
        # x = f_{i - 2}
        # y = f_{i - 1}
        temp = x + y    # Now temp = f_{i}
        x = y           # Now x = f_{i - 1}
        y = temp        # Now y = f_{i}
    return y


def fibo_it_python_style(n):
    """Fibonacci sequence (iterative computation, in a more Pythonic style)

    :param n: A nonnegative integer.

    :return: f_n, the n-th Fibonacci number, with f_0 = 0, f_1 = 1 and
    f_n = f_{n-2} + f_{n-1} for n >= 2.
    """
    # Remarque: en Python, il est possible d'affecter deux variables en même
    # temps, même si leurs anciennes valeurs sont utilisées dans le membre de
    # droite de l'affectation. Par exemple, on peut échanger les valeurs de
    # deux variables a et b ainsi:
    #   (a, b) = (b, a)
    # sans passer explicitement par une variable temporaire.
    #
    # Grâce à ce type de syntaxe, on peut récrire la fonction précédente d'une
    # manière plus agréable à lire et qui montre mieux l'esprit de
    # l'algorithme, sans se focaliser sur les détails d'implantation.
    if n == 0:
        return 0
    x = 0
    y = 1
    for i in range(2, n+1):
        # Beginning of loop :
        # x = f_{i - 2}
        # y = f_{i - 1}
        (y, x) = (x + y, y)         # ==> Voici la ligne qui change
        # End of loop :
        # x = f_{i - 1}
        # y = f_{i}
    return y


def _fibo_smart_rec_aux(n):
    """Fibonacci sequence (smart recursive computation)

    :param n: A strictly positive integer.

    :return: (f_n, f_{n-1}), where f_n is the n-th Fibonacci number,
    with f_0 = 0, f_1 = 1 and f_n = f_{n-2} + f_{n-1} for
    n >= 2.
    """
    if n == 1:
        return 1, 0
    f_n_minus_1, f_n_minus_2 = _fibo_smart_rec_aux(n - 1)
    return f_n_minus_1 + f_n_minus_2, f_n_minus_1


def fibo_smart_rec(n):
    """Fibonacci sequence (smart recursive computation)

    :param n: A nonnegative integer.

    :return: (f_n, f_{n-1}, where f_n  is the n-th Fibonacci number,
    with f_0 = 0, f_1 = 1 and f_n = f_{n-2} + f_{n-1} for
    n >= 2.
    """
    if n == 0:
        return 0
    f_n, _ = _fibo_smart_rec_aux(n)
    return f_n


if __name__ == '__main__':

    import time

    print("Sanity checks")
    print("")
    for n_test in range(10):
        print("fibo_rec(" + str(n_test) + ") = ", fibo_rec(n_test))
    print("")
    for n_test in range(10):
        print("fibo_it(" + str(n_test) + ") = ", fibo_it(n_test))
    print("")
    for n_test in range(10):
        print("fibo_it_python_style(" + str(n_test) + ") = ", fibo_it(n_test))
    print("")
    for n_test in range(10):
        print("fibo_smart_rec(" + str(n_test) + ") = ", fibo_smart_rec(n_test))

    print("")
    print("Time comparisons")
    for n_test in range(25, 36):
        print("")
        print("n_test = ", n_test)

        time1 = time.time()
        _ = fibo_rec(n_test)
        time2 = time.time()
        print("Time rec (seconds) = ", time2 - time1)

        time1 = time.time()
        _ = fibo_it(n_test)
        time2 = time.time()
        print("Time it (seconds) = ", time2 - time1)

        time1 = time.time()
        _ = fibo_smart_rec(n_test)
        time2 = time.time()
        print("Time smart rec (seconds) = ", time2 - time1)