import numpy as np
import time
import math


def classic_matrix_mult(a, b):
    """Matrices multiplication (classic method)

    :param a: A matrix (numpy 2D array).
    :param b: A matrix (numpy 2D array). a's number of columns and b's number
    of rows must be equal.

    :return: The matrix product of a and b.
    """
    # N.B. : l'énoncé demande seulement d'implanter cet algorithme pour des
    # matrices carrées, mais considérer des matrices de tailles quelconques
    # ne demande presque pas de travail supplémentaire. C'est donc ce que nous
    # faisons ici.
    (a_rows, a_cols) = a.shape
    (b_rows, b_cols) = b.shape
    if a_cols != b_rows:
        raise ValueError("Dimensions do not match")
    c = np.zeros((a_rows, b_cols))
    for i in range(a_rows):
        for j in range(b_cols):
            # We will compute c[i, j]
            for k in range(a_cols):
                c[i, j] += a[i, k] * b[k, j]
    return c


def matrix_plus(a, b):
    """Matrices sum

    :param a: A matrix (numpy 2D array).
    :param b: A matrix with same size as a.

    :return: The sum a + b.
    """
    rows, cols = a.shape
    c = np.zeros((rows, cols))
    for i in range(rows):
        for j in range(cols):
            c[i, j] = a[i, j] + b[i, j]
    return c


def matrix_minus(a, b):
    """Matrices difference

    :param a: A matrix (numpy 2D array).
    :param b: A matrix with same size as a.

    :return: The difference a - b.
    """
    rows, cols = a.shape
    c = np.zeros((rows, cols))
    for i in range(rows):
        for j in range(cols):
            c[i, j] = a[i, j] - b[i, j]
    return c


def strassen_matrix_mult(a, b):
    """Matrices multiplication (Strassen algorithm)

    :param a: A square matrix (numpy 2D array) of size 2^k for some k.
    :param b: A square matrix with same size as a.

    :return: The matrix product of a and b.
    """
    n, _ = a.shape
    if n == 1:
        return classic_matrix_mult(a, b)
    k = n // 2
    p_1 = strassen_matrix_mult(
        matrix_plus(a[0:k, 0:k], a[k:n, k:n]),
        matrix_plus(b[0:k, 0:k], b[k:n, k:n])
    )
    p_2 = strassen_matrix_mult(
        matrix_plus(a[k:n, 0:k], a[k:n, k:n]),
        b[0:k, 0:k]
    )
    p_3 = strassen_matrix_mult(
        a[0:k, 0:k],
        matrix_minus(b[0:k, k:n], b[k:n, k:n])
    )
    p_4 = strassen_matrix_mult(
        a[k:n, k:n],
        matrix_minus(b[k:n, 0:k], b[0:k, 0:k])
    )
    p_5 = strassen_matrix_mult(
        matrix_plus(a[0:k, 0:k], a[0:k, k:n]),
        b[k:n, k:n]
    )
    p_6 = strassen_matrix_mult(
        matrix_minus(a[k:n, 0:k], a[0:k, 0:k]),
        matrix_plus(b[0:k, 0:k], b[0:k, k:n])
    )
    p_7 = strassen_matrix_mult(
        matrix_minus(a[0:k, k:n], a[k:n, k:n]),
        matrix_plus(b[k:n, 0:k], b[k:n, k:n])
    )
    c = np.zeros((n, n))
    c[0:k, 0:k] = matrix_minus(matrix_plus(matrix_plus(p_1, p_4), p_7), p_5)
    c[0:k, k:n] = matrix_plus(p_3, p_5)
    c[k:n, 0:k] = matrix_plus(p_2, p_4)
    c[k:n, k:n] = matrix_minus(matrix_plus(matrix_plus(p_1, p_3), p_6), p_2)
    return c


def strassen_matrix_mult_2(a, b):
    """Matrices multiplication (Strassen algorithm, improved)

    :param a: A square matrix (numpy 2D array) of size 2^k for some k.
    :param b: A square matrix with same size as a.

    :return: The matrix product of a and b.
    """
    n, _ = a.shape
    # Difference with strassen_matrix_mult_1 : for small matrices, use the
    # classic algorithm.
    n_change_algo = 32
    if n <= n_change_algo:
        return classic_matrix_mult(a, b)
    k = n // 2
    p_1 = strassen_matrix_mult_2(
        matrix_plus(a[0:k, 0:k], a[k:n, k:n]),
        matrix_plus(b[0:k, 0:k], b[k:n, k:n])
    )
    p_2 = strassen_matrix_mult_2(
        matrix_plus(a[k:n, 0:k], a[k:n, k:n]),
        b[0:k, 0:k]
    )
    p_3 = strassen_matrix_mult_2(
        a[0:k, 0:k],
        matrix_minus(b[0:k, k:n], b[k:n, k:n])
    )
    p_4 = strassen_matrix_mult_2(
        a[k:n, k:n],
        matrix_minus(b[k:n, 0:k], b[0:k, 0:k])
    )
    p_5 = strassen_matrix_mult_2(
        matrix_plus(a[0:k, 0:k], a[0:k, k:n]),
        b[k:n, k:n]
    )
    p_6 = strassen_matrix_mult_2(
        matrix_minus(a[k:n, 0:k], a[0:k, 0:k]),
        matrix_plus(b[0:k, 0:k], b[0:k, k:n])
    )
    p_7 = strassen_matrix_mult_2(
        matrix_minus(a[0:k, k:n], a[k:n, k:n]),
        matrix_plus(b[k:n, 0:k], b[k:n, k:n])
    )
    c = np.zeros((n, n))
    c[0:k, 0:k] = matrix_minus(matrix_plus(matrix_plus(p_1, p_4), p_7), p_5)
    c[0:k, k:n] = matrix_plus(p_3, p_5)
    c[k:n, 0:k] = matrix_plus(p_2, p_4)
    c[k:n, k:n] = matrix_minus(matrix_plus(matrix_plus(p_1, p_3), p_6), p_2)
    return c


def strassen_matrix_mult_3(a, b):
    """Matrices multiplication (Strassen algorithm, any size of square matrix)

    :param a: A square matrix (numpy 2D array).
    :param b: A square matrix with same size as a.

    :return: The matrix product of a and b.
    """
    initial_size, _ = a.shape
    n = 2**math.ceil(math.log2(initial_size))
    a_extended = np.zeros((n, n))
    a_extended[0:initial_size, 0:initial_size] = a
    b_extended = np.zeros((n, n))
    b_extended[0:initial_size, 0:initial_size] = b
    c_extended = strassen_matrix_mult_2(a_extended, b_extended)
    return c_extended[0:initial_size, 0:initial_size]


if __name__ == '__main__':

    print("Sanity checks")
    a_test = np.random.rand(4, 4)
    b_test = np.random.rand(4, 4)
    classic_prod = classic_matrix_mult(a_test, b_test)
    strassen_prod = strassen_matrix_mult(a_test, b_test)
    strassen_prod_2 = strassen_matrix_mult_2(a_test, b_test)
    numpy_prod = np.dot(a_test, b_test)
    print("a_test = ")
    print(a_test)
    print("b_test = ")
    print(b_test)
    print("classic_prod = ")
    print(classic_prod)
    print("strassen_prod = ")
    print(strassen_prod)
    print("strassen_prod_2 = ")
    print(strassen_prod_2)
    print("numpy_prod = ")
    print(numpy_prod)
    print("classic_prod and numpy_prod are (almost) equal : ",
          np.allclose(classic_prod, numpy_prod))
    print("strassen_prod and numpy_prod are (almost) equal : ",
          np.allclose(strassen_prod, numpy_prod))
    print("strassen_prod_2 and numpy_prod are (almost) equal : ",
          np.allclose(strassen_prod_2, numpy_prod))

    print("")
    print("Time tests")
    n_test = 64
    print("n_test = ", n_test)
    a_test = np.random.rand(n_test, n_test)
    b_test = np.random.rand(n_test, n_test)

    time1 = time.time()
    classic_prod = classic_matrix_mult(a_test, b_test)
    time2 = time.time()
    print("Time classic (seconds) = ", time2 - time1)

    time1 = time.time()
    strassen_prod = strassen_matrix_mult(a_test, b_test)
    time2 = time.time()
    print("Time strassen (seconds) = ", time2 - time1)

    time1 = time.time()
    strassen_prod_2 = strassen_matrix_mult_2(a_test, b_test)
    time2 = time.time()
    print("Time strassen 2 (seconds) = ", time2 - time1)

    time1 = time.time()
    numpy_prod = np.dot(a_test, b_test)
    time2 = time.time()
    print("Time numpy (seconds) = ", time2 - time1)

    # Question 3.1
    # ------------
    # Complexité en pire cas :
    # Les additions sont considérées négligeables par rapport aux
    # multiplications.
    # Algorithme classique : O(n^3).
    # Algorithm de Strassen : O(n^(log_2(7))) ~= O(n^2.8074).

    # Question 3.2
    # ------------
    # Cependant, l'algorithme de Strassen est plus lent pour les petites
    # matrices. Par exemple, considérons des matrices 2 * 2. Dans l'algorithme
    # classique, on fait 8 multiplications et 4 additions. Dans l'algorithme de
    # Strassen, on ne fait que 7 multiplications mais 18 additions.

    # Question 3.3
    # ------------
    # Pour améliorer l'algorithme de Strassen, il suffit d'utiliser
    # l'algorithme classique pour les matrices de petite taille. Dans la
    # fonction strassen_matrix_mult_2, nous avons choisi une valeur
    # particulière pour n_change_algo, en-deça de laquelle on utilise
    # l'algorithme classique.
    #
    # L'idéal serait de fixer la transition à la limite empirique où
    # l'algorithme de Strassen devient plus rapide que l'algorithme rapide.
    # Mais cette limite peut dépendre de l'implantation de ces algorithmes
    # ainsi que de la machine utilisée. Ainsi, une implantation parfaitement
    # optimisée ne peut être réalisée que pour une machine donnée.

    # Question 3.4
    # ------------
    # La fonction np.dot est beaucoup plus rapide que celles que nous avons
    # implantées (en général de plusieurs ordres de grandeurs). En voici les
    # raisons.
    #
    # La fonction np.dot, que NumPy met à notre disposition dans Python, est
    # en fait directement programmée en C, qui est un langage :
    # 1. Compilé. Le code est rédigé en langage C (compréhensible par un
    # humain) puis "compilé", c'est-à-dire transformé en fichier exécutable
    # (directement compréhensible par la machine). Quand on veut utiliser le
    # programme, on lance ce fichier exécutable (et non le code source).
    # 2. De "bas niveau", c'est-à-dire où les instructions qu'on écrit sont
    # proches de ce que fait la machine en pratique.
    #
    # En revanche, nous avons écrit nos fonctions en Python, qui est un
    # langage :
    # 1. Interprété. Le code est rédigé en Python (compréhensible par un
    # humain) et il reste en Python. Quand on veut utiliser le programme, on
    # exécute l'interpréteur Python sur ce code. La machine a donc un double
    # travail : traduire le langage humain en langage machine, et effectuer
    # le travail demandé.
    # 2. De "haut niveau", c'est-à-dire que les instructions qu'on écrit
    # correspondent davantage à l'idée qu'on a de l'algorithme, en
    # s'embarrassant moins de détails techniques (types des variables, gestion
    # de la mémoire physique sur la machine, etc.).
    #
    # Ainsi, le Python est globalement plus commode à écrire que le C, mais
    # aussi plus lent à l'exécution. Python est donc parfait pour des
    # applications où il est important d'économiser le temps de développement
    # et où l'on souhaite facilement pouvoir travailler en équipe, étendre le
    # code, ajouter de nouvelles fonctionalités. À l'inverse, C et ses
    # variantes sont privilégiés par les programmeurs qui souhaitent avoir des
    # performances optimisées au moment de l'exécution.
    #
    # Le package NumPy offre une passerelle entre ces deux mondes : pour des
    # opérations courantes et potentiellement coûteuses en temps comme la
    # multiplication de matrices, il met à disposition du programmateur
    # Python des performances comparables à celles du C.

    print("")
    print("Square matrices with any size (question 4.1)")
    print("Sanity checks")
    a_test = np.random.rand(5, 5)
    b_test = np.random.rand(5, 5)
    classic_prod = classic_matrix_mult(a_test, b_test)
    strassen_prod_3 = strassen_matrix_mult_3(a_test, b_test)
    numpy_prod = np.dot(a_test, b_test)
    print("a_test = ")
    print(a_test)
    print("b_test = ")
    print(b_test)
    print("classic_prod = ")
    print(classic_prod)
    print("strassen_prod_3 = ")
    print(strassen_prod_3)
    print("numpy_prod = ")
    print(numpy_prod)
    print("classic_prod and numpy_prod are (almost) equal : ",
          np.allclose(classic_prod, numpy_prod))
    print("strassen_prod_3 and numpy_prod are (almost) equal : ",
          np.allclose(strassen_prod_3, numpy_prod))
