# -*- coding: utf-8 -*-
class Tree:

    def __init__(self, name='', children=None):
        """A (rooted) tree.

        :param name: A String. The name of the root.
        :param children: A list of trees.
        """
        self.name = name
        if not children:
            self.children = []
        else:
            self.children = children

    def height(self):
        """Height.

        :return: An integer. The height of the tree. It is 0 for a leave;
        otherwise, it is the maximum height of children trees, plus one.
        """
        if not self.children:
            return 0
        max_child_height = 0
        for child in self.children:
            max_child_height = max(max_child_height, child.height())
        return 1 + max_child_height

    @property
    def height_advanced(self):
        """Height (as a property and using a list comprehension).

        :return: An integer. The height of the tree. It is 0 for a leave;
        otherwise, it is the maximum height of children trees, plus one.

        Technical notes for students:
        - Here, 'height_advanced' is defined with the decorator @property,
        which makes it considered as an attribute. For example, the height
        of tree gaia can be called by gaia.height_advanced (without
        parenthesis because it is considered as an attribute). In contrast,
        consider the syntax gaia.height() (with parenthesis because it is a
        method).
        - The height can be implemented with a 'for' loop as done in method
        'height' above, but it is much more elegant and pythonic with a
        'list comprehension', as is implemented here (see
        docs.python.org/3/tutorial/datastructures.html#list-comprehensions).
        """
        if not self.children:
            return 0
        return 1 + max([child.height_advanced for child in self.children])

    def display(self, indent=""):
        """Print tree in a human-friendly way.
        """
        print(indent + self.name)
        for child in self.children:
            child.display(indent + "    ")

    def size(self):
        """Size.

        :return: An integer. The size of the tree, i.e. the total number of
        nodes.
        """
        if not self.children:
            return 1
        return 1 + sum([child.size() for child in self.children])

    def nb_leaves(self):
        """Number of leaves

        :return: An integer. The number of leaves, i.e. the number of nodes
        without child.
        """
        if not self.children:
            return 1
        return sum([child.nb_leaves() for child in self.children])

    def search_dfs(self, name):
        """Search.

        :param name: A string.

        :return: A Tree. The subtree rooted at the node whose name is equal to
        parameter name. If name is not found, then return None. If name is
        present more than one time, then return its first occurrence by a
        Depth First Search.
        """
        if self.name == name:
            return self
        for child in self.children:
            result = child.search_dfs(name)
            if result is not None:
                return result
        return None

    def display_bfs(self):
        """Display Breadth-First Search.

        Print the name of the root, then the names of all nodes with depth
        1, then the names of all nodes with depth 2, etc.
        """
        fifo = [self]
        while fifo:
            element = fifo.pop(0)
            print(element.name)
            fifo.extend(element.children)


class MythologyTree(Tree):

    def __init__(self, name='', children=None, status=None, gender=None):
        """A simplified family tree.

        :param name: A String. The name of the root.
        :param children: A list of trees.
        :param status: A String. The status of the root (usually Deity,
        Titan, God, etc.).
        :param gender: A String. The gender of the root (usually F or M).
        """
        super().__init__(name, children)
        self.status = status
        self.gender = gender


if __name__ == "__main__":

    persephone = MythologyTree(name="Persephone", status="God", gender="F")
    demeter = MythologyTree(name="Demeter", status="God", gender="F",
                            children=[persephone])
    zeus = MythologyTree(name="Zeus", status="God", gender="M")
    cronos = MythologyTree(name="Cronos", status="Titan", gender="M",
                           children=[zeus, demeter])
    hyperion = MythologyTree(name="Hyperion", status="Titan", gender="M")
    gaia = MythologyTree(name="Gaia", status="Deity", gender="F",
                         children=[cronos, hyperion])

    gaia.display()

    print("")
    print("Gaia's height:", gaia.height())
    print("Cronos' height:", cronos.height())
    print("Hyperion's height:", hyperion.height())

    print("")
    print("Gaia's height (advanced version):", gaia.height())
    print("Cronos' height (advanced version):", cronos.height())
    print("Hyperion's height (advanced version):", hyperion.height())

    print("")
    print("Gaia's size:", gaia.size())
    print("Cronos' size:", cronos.size())

    print("")
    print("Gaia's number of leaves:", gaia.nb_leaves())
    print("Cronos' number of leaves:", cronos.nb_leaves())

    print("")
    print("Information about Demeter: ")
    search_result = gaia.search_dfs("Demeter")
    print("Status:", search_result.status)
    print("Gender:", search_result.gender)

    print("")
    print("BFS display:")
    gaia.display_bfs()